# -*- coding: utf-8 -*-

import asyncio
import binascii
import json
import logging
import time
import typing

import requests

from paradox.config import config as cfg
from paradox.lib import stun
from paradox.lib.crypto import encrypt
from paradox.parsers.paradox_ip_messages import ip_message, ip_payload_connect_response
from .connection import Connection
from .protocols import SerialConnectionProtocol, IPConnectionProtocol

logger = logging.getLogger('PAI').getChild(__name__)


class IPConnection(Connection):
    def __init__(self, on_message: typing.Callable[[bytes], None], host='127.0.0.1', port=10000, password=None):
        super(IPConnection, self).__init__(on_message=on_message)
        if isinstance(password, str):
            password = password.encode()
        self.password = password
        self.key = password
        self.host = host
        self.port = port
        self.site_info = None
        self.module = None
        self.connection_timestamp = 0

        self.stun_control = None
        self.stun_tunnel = None
        self.connection = None

    def on_connection_lost(self):
        logger.error('Connection to panel was lost')
        self.close()

    def close(self):
        logger.info('Closing IP Connection')

        self.connected = False
        self.connection_timestamp = 0
        
        if self.stun_control:
            try:
                self.stun_control.close()
                self.stun_control = None
            except Exception as e:
                logger.exception("stun_control socket close failed")
        if self.stun_tunnel:
            try:
                self.stun_tunnel.close()
                self.stun_tunnel = None
            except Exception as e:
                logger.exception("stun_control socket close failed")
        if self.connection:
            try:
                self.connection.close()
            except Exception as e:
                logger.exception("connection socket close failed")
            self.connection = None

    def make_protocol(self):
        if cfg.IP_CONNECTION_BARE:
            return SerialConnectionProtocol(self.on_message, self.on_bare_connection_open, self.on_connection_lost)
        else:
            return IPConnectionProtocol(self.on_message, self.on_connection_lost, self.key)

    def on_bare_connection_open(self):
        logger.info('Serial port open')
        self.connected = True

    async def connect(self):
        tries = 1
        max_tries = 3

        while tries <= max_tries:

            if cfg.IP_CONNECTION_SITEID is not None and cfg.IP_CONNECTION_EMAIL is not None:
                try:
                    r = await self.connect_to_site()

                    if r and self.site_info is not None:
                        if await self.connect_to_module():
                            return True
                except Exception:
                    logger.exception('Try %d/%d. Unable to connect to SITE ID' % (tries, max_tries))
            else:
                try:
                    logger.info("Connecting to IP module. Try %d/%d" % (tries, max_tries))

                    _, self.connection = await self.loop.create_connection(
                        self.make_protocol,
                        host=self.host,
                        port=self.port
                    )
                    if cfg.IP_CONNECTION_BARE:
                        return True

                    if await self.connect_to_module():
                        return True
                except OSError as e:
                    logger.error('Connect to IP Module failed (try %d/%d): %s' % (tries, max_tries, str(e)))
                except Exception:
                    logger.exception("Unable to connect to IP Module (try %d/%d)" % (tries, max_tries))

            tries += 1

        return False

    async def connect_to_site(self):
        self.connection_timestamp = 0
        logger.info("Connecting to Site: {}".format(cfg.IP_CONNECTION_SITEID))
        if self.site_info is None:
            self.site_info = self.get_site_info(siteid=cfg.IP_CONNECTION_SITEID, email=cfg.IP_CONNECTION_EMAIL)

        if self.site_info is None:
            logger.error("Unable to get site info")
            return False
        try:
            # xoraddr = binascii.unhexlify(self.site_info['site'][0]['module'][0]['xoraddr'])
            if self.site_info is None:
                logger.error("Unable to get site info")
                return False

            self.module = None

            logger.debug("Site Info: {}".format(json.dumps(self.site_info, indent=4)))

            if cfg.IP_CONNECTION_PANEL_SERIAL is not None:
                for site in self.site_info['site']:
                    for module in site:
                        logger.debug("Found module with panel serial: {}".format(module['panelSerial']))
                        if module['panelSerial'] == cfg.IP_CONNECTION_PANEL_SERIAL:
                            self.module = module
                            break

                    if self.module is not None:
                        break
            else:
                self.module = self.site_info['site'][0]['module'][0]  # Use first

            if self.module is None:
                self.site_info = None  # Reset state
                logger.error("Unable to find module with desired panel serial")
                return False

            xoraddr = binascii.unhexlify(self.module['xoraddr'])

            stun_host = 'turn.paradoxmyhome.com'

            logger.debug("STUN TCP Change Request")
            self.stun_control = stun.StunClient(stun_host)
            self.stun_control.send_tcp_change_request()
            stun_r = self.stun_control.receive_response()
            if stun.is_error(stun_r):
                logger.error(stun.get_error(stun_r))
                return False

            logger.debug("STUN TCP Binding Request")
            self.stun_control.send_binding_request()
            stun_r = self.stun_control.receive_response()
            if stun.is_error(stun_r):
                logger.error(stun.get_error(stun_r))
                return False

            logger.debug("STUN Connect Request")
            self.stun_control.send_connect_request(xoraddr=xoraddr)
            stun_r = self.stun_control.receive_response()
            if stun.is_error(stun_r):
                logger.error(stun.get_error(stun_r))
                return False

            self.connection_timestamp = time.time()

            connection_id = stun_r[0]['attr_body']
            raddr = self.stun_control.sock.getpeername()

            logger.debug("STUN Connection Bind Request")
            self.stun_tunnel = stun.StunClient(host=raddr[0], port=raddr[1])
            self.stun_tunnel.send_connection_bind_request(binascii.unhexlify(connection_id))
            stun_r = self.stun_tunnel.receive_response()
            if stun.is_error(stun_r):
                logger.error(stun.get_error(stun_r))
                return False

            _, self.connection = await self.loop.create_connection(self.make_protocol, sock=self.stun_tunnel.sock)
            logger.info("Connected to Site: {}".format(cfg.IP_CONNECTION_SITEID))
        except Exception:
            logger.exception("Unable to negotiate connection to site")
            return False

        return True

    async def connect_to_module(self):
        try:
            logger.info("Authenticating with IP Module")

            self.key = self.password  # first request is with initial password, next with generated by panel key

            self.connection.key = self.password
            payload = encrypt(self.password, self.key)

            msg = ip_message.build(
                dict(header=dict(length=len(self.key), unknown0=0x03, flags=0x09, command=0xf0, unknown1=0, encrypt=1),
                     payload=payload))
            self.connection.send_raw(msg)
            message_payload = await self.wait_for_message(raw=True)

            response = ip_payload_connect_response.parse(message_payload)

            if response.login_status != 'success':
                logger.error("Error connecting to IP Module. Wrong IP Module password?")
                return False

            logger.info("Authentication Success. IP Module version {:02x}, firmware: {}.{}, serial: {}".format(
                response.hardware_version,
                response.ip_firmware_major,
                response.ip_firmware_minor,
                binascii.hexlify(response.ip_module_serial).decode('utf-8')))

            self.key = response.key
            self.connection.key = response.key

            # F2
            logger.debug("Sending F2")
            msg = ip_message.build(
                dict(header=dict(length=0, unknown0=0x03, flags=0x09, command=0xf2, unknown1=0, encrypt=1),
                     payload=encrypt(b'', self.key)))
            self.connection.send_raw(msg)
            message_payload = await self.wait_for_message(raw=True)
            logger.debug("F2 answer: {}".format(binascii.hexlify(message_payload)))

            # # F4
            # logger.debug("Sending F4")
            # msg = binascii.unhexlify('aa00000309f400000001eeeeeeee0000')
            # self.connection.send_raw(msg)
            # message_payload = await self.wait_for_message(raw=True)
            #
            # logger.debug("F4 answer: {}".format(binascii.hexlify(message_payload)))

            # F3
            logger.debug("Sending F3")
            msg = ip_message.build(
                dict(header=dict(length=0, unknown0=0x03, flags=0x09, command=0xf3, unknown1=0, encrypt=1),
                     payload=encrypt(b'', self.key)))
            self.connection.send_raw(msg)
            message_payload = await self.wait_for_message(raw=True)

            # logger.debug("F3 answer: {}".format(binascii.hexlify(message_payload)))

            # F8
            logger.debug("Sending F8")
            payload = binascii.unhexlify('0a500080000000000000000000000000000000000000000000000000000000000000000000d0')
            payload_len = len(payload)
            payload = encrypt(payload, self.key)
            msg = ip_message.build(
                dict(header=dict(length=payload_len, unknown0=0x03, flags=0x09, command=0xf8, unknown1=0, encrypt=1),
                     payload=payload))
            self.connection.send_raw(msg)
            message_payload = await self.wait_for_message(raw=True)
            logger.debug("F8 answer: {}".format(binascii.hexlify(message_payload)))

            logger.info("Session Established with IP Module")

            self.connected = True
        except asyncio.TimeoutError:
            self.connected = False
            logger.error(
                "Unable to establish session with IP Module. Timeout. Only one connection at a time is allowed."
            )
        except Exception as e:
            self.connected = False
            logger.exception("Unable to establish session with IP Module")

        return self.connected

    def write(self, data: bytes):
        """Write data to socket"""

        if not self.refresh_stun():
            raise ConnectionError('Failed to refresh STUN')

        return super(IPConnection, self).write(data)

    @staticmethod
    def get_site_info(email, siteid):
        logger.info("Getting site info")
        URL = "https://api.insightgoldatpmh.com/v1/site"

        headers = {'User-Agent': 'Mozilla/3.0 (compatible; Indy Library)', 'Accept-Encoding': 'identity',
                   'Accept': 'text/html, */*'}

        tries = 5
        while tries > 0:
            req = requests.get(URL, headers=headers, params={'email': email, 'name': siteid})
            if req.status_code == 200:
                return req.json()

            logger.warning("Unable to get site info. Retrying...")
            tries -= 1
            time.sleep(5)

        return None

    def refresh_stun(self):
        if self.site_info is None or self.connection_timestamp == 0:
            return True

        try:
            # Refresh session if required
            if time.time() - self.connection_timestamp >= 500:
                logger.info("Refreshing session")
                self.stun_control.send_refresh_request()
                stun_r = self.stun_control.receive_response()
                if stun.is_error(stun_r):
                    logger.error(stun.get_error(stun_r))
                    self.connected = False
                    return False

                self.connection_timestamp = time.time()

            return True
        except Exception:
            logger.exception("Session refresh")
            return False



import asyncio
import logging
import os
import socket
import time
import typing
from enum import Enum

from paho.mqtt.client import Client, MQTT_ERR_SUCCESS

from paradox.config import config as cfg
from paradox.data.enums import RunState
from paradox.interfaces import ThreadQueueInterface
from paradox.lib import ps

logger = logging.getLogger('PAI').getChild(__name__)

ELEMENT_TOPIC_MAP = dict(partition=cfg.MQTT_PARTITION_TOPIC, zone=cfg.MQTT_ZONE_TOPIC,
                         output=cfg.MQTT_OUTPUT_TOPIC, repeater=cfg.MQTT_REPEATER_TOPIC,
                         bus=cfg.MQTT_BUS_TOPIC, keypad=cfg.MQTT_KEYPAD_TOPIC,
                         system=cfg.MQTT_SYSTEM_TOPIC, user=cfg.MQTT_USER_TOPIC)


class ConnectionState(Enum):
    NEW = 0
    CONNECTING = 1
    CONNECTED = 2
    DISCONNECTING = 3


RUN_STATE_2_PAYLOAD = {
    RunState.ERROR: "error",
    RunState.INIT: "initializing",
    RunState.PAUSE: "paused",
    RunState.RUN: "online",
    RunState.STOP: "stopped",
}


class MQTTConnection(Client):
    _instance = None

    @classmethod
    def get_instance(cls) -> 'MQTTConnection':
        if cls._instance is None:
            cls._instance = MQTTConnection()

        return cls._instance

    def __init__(self):
        super(MQTTConnection, self).__init__("paradox_mqtt/{}".format(os.urandom(8).hex()))
        self._last_run_state = "unknown"
        self.run_status_topic = '{}/{}/{}'.format(cfg.MQTT_BASE_TOPIC, cfg.MQTT_INTERFACE_TOPIC, 'run_status')
        self.availability_topic = '{}/{}/{}'.format(cfg.MQTT_BASE_TOPIC, cfg.MQTT_INTERFACE_TOPIC, 'availability')
        self.on_connect = self._on_connect_cb
        self.on_disconnect = self._on_disconnect_cb
        self.state = ConnectionState.NEW
        # self.on_subscribe = lambda client, userdata, mid, granted_qos: logger.debug("Subscribed: %s" %(mid))
        # self.on_message = lambda client, userdata, message: logger.debug("Message received: %s" % str(message))
        # self.on_publish = lambda client, userdata, mid: logger.debug("Message published: %s" % str(mid))

        ps.subscribe(self.on_run_state_change, "run-state")

        self.registrars = []

        if cfg.MQTT_USERNAME is not None and cfg.MQTT_PASSWORD is not None:
            self.username_pw_set(username=cfg.MQTT_USERNAME, password=cfg.MQTT_PASSWORD)

        self.will_set(self.availability_topic, 'offline', 0, retain=True)

    def on_run_state_change(self, state: RunState):
        v = RUN_STATE_2_PAYLOAD.get(state, "unknown")
        self._report_run_state(v)

    def start(self):
        if self.state == ConnectionState.NEW:
            self.loop_start()

            # TODO: Some initial connection retry mechanism required
            try:
                self.connect_async(
                    host=cfg.MQTT_HOST, port=cfg.MQTT_PORT, keepalive=cfg.MQTT_KEEPALIVE,
                    bind_address=cfg.MQTT_BIND_ADDRESS, bind_port=cfg.MQTT_BIND_PORT
                )

                self.state = ConnectionState.CONNECTING

                logger.info("MQTT loop started")
            except socket.gaierror:
                logger.exception("Failed to connect to MQTT (%s:%d)", cfg.MQTT_HOST, cfg.MQTT_PORT)

    def stop(self):
        if self.state in [ConnectionState.CONNECTING, ConnectionState.CONNECTED]:
            self.disconnect()
            self.loop_stop()
            logger.info("MQTT loop stopped")

    def publish(self, topic, payload=None, *args, **kwargs):
        logger.debug("MQTT: {}={}".format(topic, payload))

        super(MQTTConnection, self).publish(topic, payload, *args, **kwargs)

    def _call_registars(self, method, *args, **kwargs):
        for r in self.registrars:
            try:
                if hasattr(r, method) and isinstance(getattr(r, method), typing.Callable):
                    getattr(r, method)(*args, **kwargs)
            except Exception as e:
                logger.exception('Failed to call "%s" on "%s"', method, r.__class__.__name__)

    def register(self, cls):
        self.registrars.append(cls)

    @property
    def connected(self):
        return self.state == ConnectionState.CONNECTED

    def _report_run_state(self, status):
        self._last_run_state = status
        self.publish(self.run_status_topic, status, 0, retain=True)
        self.publish(self.availability_topic, "online" if status in ["online", "paused"] else "offline", 0, retain=True)

    def _on_connect_cb(self, client, userdata, flags, result):
        if result == MQTT_ERR_SUCCESS:
            logger.info("MQTT Broker Connected")
            self.state = ConnectionState.CONNECTED
            self._report_run_state(self._last_run_state)
            self._call_registars("on_connect", client, userdata, flags, result)
        else:
            logger.error("Failed to connect to MQTT. Code: %d" % result)

    def _on_disconnect_cb(self, client, userdata, rc):
        if rc == MQTT_ERR_SUCCESS:
            logger.info("MQTT Broker Disconnected")
        else:
            logger.error("MQTT Broker unexpectedly disconnected. Code: %d", rc)

        self.state = ConnectionState.NEW
        self._call_registars("on_disconnect", client, userdata, rc)

    def disconnect(self, reasoncode=None, properties=None):
        self.state = ConnectionState.DISCONNECTING
        self._report_run_state('offline')
        super(MQTTConnection, self).disconnect()


class AbstractMQTTInterface(ThreadQueueInterface):
    """Interface Class using MQTT"""
    def __init__(self, alarm):
        super().__init__(alarm)

        self.mqtt = MQTTConnection.get_instance()
        self.mqtt.register(self)
        logger.debug("Registars: %d", len(self.mqtt.registrars))

        self.republish_cache = {}

    def start(self):
        super().start()
        self.mqtt.start()

    def stop(self):
        """ Stops the MQTT Interface Thread"""
        self.mqtt.stop()
        super().stop()
        self.republish_task.cancel()
        self.loop.call_soon_threadsafe(self.loop.stop)

    async def republish_loop(self):
        while True:
            await asyncio.sleep(cfg.MQTT_REPUBLISH_INTERVAL)
            trigger = time.time() - cfg.MQTT_REPUBLISH_INTERVAL

            for k, v in filter(lambda f: f[1].get("last_publish") <= trigger, self.republish_cache.items()):
                self.publish(k, v['value'], v['qos'], v['retain'])

    def _run(self):
        super(AbstractMQTTInterface, self)._run()

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.republish_task = self.loop.create_task(self.republish_loop())

        self.loop.run_forever()
        self.loop.close()

    def publish(self, topic, value, qos, retain):
        self.republish_cache[topic] = {'value': value, 'qos': qos, 'retain': retain, 'last_publish': time.time()}
        self.loop.call_soon_threadsafe(self.mqtt.publish, topic, value, qos, retain)

    def subscribe_callback(self, sub, callback: typing.Callable):
        self.mqtt.message_callback_add(sub, callback)
        self.mqtt.subscribe(sub)

    def on_disconnect(self, client, userdata, rc):
        """ Called from MQTT connection """
        pass

    def on_connect(self, client, userdata, flags, result):
        """ Called from MQTT connection """
        pass



import os
import sys
import logging


class Config(object):
    DEFAULTS = {
        "LOGGING_LEVEL_CONSOLE": logging.INFO,  # See documentation of Logging package
        "LOGGING_LEVEL_FILE": logging.ERROR,
        "LOGGING_FILE": ('/var/log/paradox.log', [type(None), str]),  # or set to file path : '/var/log/paradox.log'
        "LOGGING_FILE_MAX_SIZE": (10, int, (0, 0xFFFFFFFF)),          # Max log file size in MB
        "LOGGING_FILE_MAX_FILES": (2, int, (0, 0xFFFFFFFF)),          # Max old log files to keep
        "LOGGING_DUMP_PACKETS": False,          # Dump RAW Packets to log
        "LOGGING_DUMP_MESSAGES": False,         # Dump Messages to log
        "LOGGING_DUMP_STATUS": False,           # Dump Status to log
        "LOGGING_DUMP_EVENTS": False,           # Dump Event details to log

        # Development
        "DEVELOPMENT_DUMP_MEMORY": False,

        # Connection Type
        "CONNECTION_TYPE": ('Serial', str, ['IP', 'Serial']),         # Serial or IP

        # Serial Connection Details
        "SERIAL_PORT": '/dev/ttyS1',                                  # Pathname of the Serial Port
        "SERIAL_BAUD": 9600,                                          # Baud rate of the Serial Port. Use 38400(default setting) or 57600 for EVO

        # IP Connection Details
        "IP_CONNECTION_HOST": '127.0.0.1',                            # IP Module address when using direct IP Connection
        "IP_CONNECTION_PORT": (10000, int, (1, 65535)),               # IP Module port when using direct IP Connection
        "IP_CONNECTION_PASSWORD": ('paradox', [str, bytes, type(None)]),  # IP Module password. "paradox" is default.
        "IP_CONNECTION_SITEID": (None, [str, type(None)]),            # If defined, connection will be made through this method.
        "IP_CONNECTION_EMAIL": (None, [str, type(None)]),             # Email registered in the site
        "IP_CONNECTION_PANEL_SERIAL": (None, [str, type(None)]),      # Serial number to be used in multi-panel sites. None for first
        "IP_CONNECTION_BARE": False,                                  # IP endpoint connects directly to panel. Used for Serial Tunnels over TCP
        # Paradox
        "KEEP_ALIVE_INTERVAL": 10,        # Interval between status updates

        "LIMITS": {},  # By default all zones will be monitored

        "LABEL_ENCODING": "utf-8",											  # Encoding to use when decoding labels. See https://docs.python.org/3/library/codecs.html#standard-encodings
        "LABEL_REFRESH_INTERVAL": (15 * 60, int, (0, 0xFFFFFFFF)),            # Interval between refresh of labels
        "OUTPUT_PULSE_DURATION": (1, float, (0, 0xFFFFFFFF)),                 # Duration of a PGM pulse in seconds
        "SYNC_TIME": False,                                                   # Update panel time
        "PASSWORD": (None, [int, str, bytes, type(None)]),                    # PC Password. Set to None if Panel has no Password. In Babyware: Right click on your panel -> Properties -> PC Communication (Babyware) -> PC Communication (Babyware) Tab.

        "POWER_UPDATE_INTERVAL": (60, int, (0, 0xFFFFFFFF)),     # Interval between updates of the Power voltages
        "PUSH_POWER_UPDATE_WITHOUT_CHANGE": True,  # Always notify interfaces of power changes
        "PUSH_UPDATE_WITHOUT_CHANGE": False,       # Always notify interfaces of all changes

        # MQTT
        "MQTT_ENABLE": False,                        # Enable MQTT Interface
        "MQTT_HOST": '127.0.0.1',                    # Hostname or address
        "MQTT_PORT": (1883, int, (1, 65535)),        # TCP Port
        "MQTT_KEEPALIVE": (60, int, (1, 3600)),      # Keep alive
        "MQTT_USERNAME": (None, [str, type(None)]),  # MQTT Username for authentication
        "MQTT_PASSWORD": (None, [str, type(None)]),  # MQTT Password
        "MQTT_RETAIN": True,                         # Publish messages with Retain
        "MQTT_BIND_ADDRESS": '127.0.0.1',            # MQTT Bind address
        "MQTT_BIND_PORT": 0,                         # MQTT Bind port
        "MQTT_REPUBLISH_INTERVAL": (60 * 60 * 12, int, (60, 0xFFFFFFFF)),    # Interval for republishing all data
        "MQTT_HOMEASSISTANT_AUTODISCOVERY_ENABLE": False,

        # MQTT Topics
        "MQTT_BASE_TOPIC": 'paradox',             # Root of all topics
        "MQTT_ZONE_TOPIC": 'zones',               # Base for zone states
        "MQTT_PARTITION_TOPIC": 'partitions',     # Base for partition states
        "MQTT_BUS_TOPIC": 'buses',                # Base for buses states
        "MQTT_SYSTEM_TOPIC": 'system',            # Base for panel states
        "MQTT_REPEATER_TOPIC": 'repeaters',       # Base for repeater states
        "MQTT_USER_TOPIC": 'users',               # Base for user states
        "MQTT_EVENTS_TOPIC": 'events',            # Base for events
        "MQTT_CONTROL_TOPIC": 'control',          # Base for control of othe elements (ROOT/CONTROL/TYPE)
        "MQTT_DEFINITION_TOPIC": 'definitions',    # Base for definitions
        "MQTT_HOMEASSISTANT_CONTROL_TOPIC": 'hass_control',
        "MQTT_HOMEASSISTANT_DISCOVERY_PREFIX": 'homeassistant',
        "MQTT_OUTPUT_TOPIC": 'outputs',
        "MQTT_KEYPAD_TOPIC": 'keypads',
        "MQTT_STATES_TOPIC": 'states',
        "MQTT_RAW_TOPIC": 'raw',
        "MQTT_NOTIFICATIONS_TOPIC": 'notifications',
        "MQTT_PUBLISH_RAW_EVENTS": True,
        "MQTT_INTERFACE_TOPIC": 'interface',
        "MQTT_TOGGLE_CODES": {},
        "MQTT_USE_NUMERIC_STATES": False,         # use 0 and 1 instead of True and False
        "MQTT_DASH_PUBLISH": False,
        "MQTT_DASH_TOPIC": "metrics/exchange/pai",
        "MQTT_DASH_TEMPLATE": "/etc/pai/mqtt_dash.txt",

        # Interfaces text command alias
        "COMMAND_ALIAS": {                       # alias for commands through text based interfaces
            'arm': 'partition all arm',
            'disarm': 'partition all disarm'
        },

        # MQTT command aliases
        "MQTT_COMMAND_ALIAS": {
            # For homebridge
            'armed_home': 'arm_stay',
            'armed_night': 'arm_sleep',
            'armed_away': 'arm',
            'disarmed': 'disarm'
        },

        # Pushbullet
        "PUSHBULLET_ENABLE": False,
        "PUSHBULLET_KEY": '',                     # Authentication key used for Pushbullet
        "PUSHBULLET_CONTACTS": [],                # Pushbullet user identifiers for notifications and interaction
        "PUSHBULLET_IGNORE_EVENTS": [],           # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "PUSHBULLET_ALLOW_EVENTS": [],            # Same as before but as a white list. Default is use EVENT_FILTERS
        "PUSHBULLET_EVENT_FILTERS": [             # list of tags, property changes to include or exclude. See event.py for tag list
            'alarm,-restore',
            'trouble,-clock',
            'tamper'
        ],
        "PUSHBULLET_MIN_EVENT_LEVEL": ('INFO', str, ['DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL']),

        # Pushover
        "PUSHOVER_ENABLE": False,
        "PUSHOVER_KEY": '',                       # Application token for Pushover
        "PUSHOVER_BROADCAST_KEYS": {              # Pushover user or group keys to broadcast notifications to
            #    '<user_key>': '*'                # value can be '*' or comma separated list of device names
        },
        "PUSHOVER_IGNORE_EVENTS": [],             # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "PUSHOVER_ALLOW_EVENTS": [],              # Same as before but as a white list. Default is use EVENT_FILTERS
        "PUSHOVER_EVENT_FILTERS": [               # list of tags, property changes to include or exclude. See event.py for tag list
            'alarm,-restore',
            'trouble,-clock',
            'tamper'
        ],
        "PUSHOVER_MIN_EVENT_LEVEL": ('INFO', str, ['DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL']),

        # Signal
        "SIGNAL_ENABLE": False,
        "SIGNAL_CONTACTS": [],                  # Contacts that are allowed to control the panel and receive notifications through Signal
        "SIGNAL_IGNORE_EVENTS": [],             # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "SIGNAL_ALLOW_EVENTS": [],              # Same as before but as a white list. Default is use EVENT_FILTERS
        "SIGNAL_EVENT_FILTERS": [               # list of tags, property changes to include or exclude. See event.py for tag list
            'alarm,-restore',
            'trouble,-clock',
            'tamper'
        ],
        "SIGNAL_MIN_EVENT_LEVEL": ('INFO', str, ['DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL']),

        # GSM
        "GSM_ENABLE": False,
        "GSM_MODEM_BAUDRATE": (115200, int, (9600, 115200)),    # Baudrate of the GSM modem
        "GSM_MODEM_PORT": '',                     # Pathname of the serial port
        "GSM_CONTACTS": [],                       # Contacts that are allowed to control the panel and receive notifications through SMS
        "GSM_IGNORE_EVENTS": [],                  # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "GSM_ALLOW_EVENTS": [],                   # Same as before but as a white list. Default is use EVENT_FILTERS
        "GSM_EVENT_FILTERS": [                    # list of tags, property changes to include or exclude. See event.py for tag list
            'zone,alarm,trigger'
        ],
        "GSM_MIN_EVENT_LEVEL": ('CRITICAL', str, ['DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL']),

        # IP Socket Interface
        "IP_INTERFACE_ENABLE": False,
        "IP_INTERFACE_BIND_ADDRESS": '0.0.0.0',
        "IP_INTERFACE_BIND_PORT": (10000, int, (1, 65535)),
        "IP_INTERFACE_PASSWORD": (b'paradox', [bytes, type(None)]),

        # Dummy Interface for testing
        "DUMMY_INTERFACE_ENABLE": False,
        "DUMMY_EVENT_FILTERS": [
            'alarm-restore',
            'trouble-clock',
            'tamper'
            'arm',
            'disarm'
        ],
        "DUMMY_MIN_EVENT_LEVEL": ('DEBUG', str, ['DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'])
    }

    CONFIG_LOADED = False
    CONFIG_FILE_LOCATION = None,

    def __dir__(self):
        return list(self.DEFAULTS.keys()) + list(self.__class__.__dict__) + dir(super(Config, self))

    def load(self, alt_location=None):
        self.CONFIG_LOADED = False

        env_config_path = os.environ.get('PAI_CONFIG_FILE')

        if alt_location is not None:
            locations = [alt_location]
        elif env_config_path:
            locations = [env_config_path]
        else:
            locations = ['/etc/pai/pai.conf',
                         '/usr/local/etc/pai/pai.conf',
                         '~/.local/etc/pai.conf',
                         'pai.conf'
                         ]

        for location in locations:
            location = os.path.expanduser(location)
            if os.path.exists(location) and os.path.isfile(location):
                self.CONFIG_FILE_LOCATION = location
                break
        else:
            err = "ERROR: Could not find configuration file. Tried: {}".format(locations)
            sys.stderr.write(err+'\n')
            raise(Exception(err))

        sys.stdout.write("Attempting to load configuration from %s\n" % self.CONFIG_FILE_LOCATION)

        entries = {}
        conf_extension = os.path.splitext(self.CONFIG_FILE_LOCATION)[1]
        if conf_extension in ['.conf', '.py']:
            with open(self.CONFIG_FILE_LOCATION) as f:
                exec(f.read(), None, entries)
        elif conf_extension in ['.json']:
            import json
            with open(self.CONFIG_FILE_LOCATION) as f:
                entries = json.load(f)
        elif conf_extension in ['.yaml']:
            import yaml
            with open(self.CONFIG_FILE_LOCATION) as f:
                entries = yaml.safe_load(f)
        else:
            err = "ERROR: Unsupported configuration file type"
            sys.stderr.write(err+'\n')
            raise(Exception(err))

        # Updates values from env variables
        for args in os.environ:
            if not args.startswith('PAI_') or len(args) < 5:
                continue
            opt = args[4:]
            if opt in self.DEFAULTS:
                v = os.environ[args]
                if v.isdigit():
                    v = int(v)

                entries[opt] = v

        # Reset defaults
        for k, v in self.DEFAULTS.items():
            if isinstance(v, tuple):
                v = v[0]

            setattr(self, k, v)

        # Set values
        for k, v in entries.items():
            if k[0].isupper() and k in self.DEFAULTS:
                default = self.DEFAULTS.get(k)
                
                if isinstance(default, tuple) and 2 <= len(default) <= 3:
                    default_type = default[1]

                    if not isinstance(default_type, list):
                        default_type = [default_type]

                else:
                    default_type = [type(default)]

                if float in default_type and not int in default_type:
                    default_type.append(int)
                if int in default_type and not float in default_type:
                    default_type.append(float)

                if type(v) in default_type:
                    setattr(self, k, v)
                else:
                    err = "Error parsing configuration: Invalid value type {} for config argument {}. Allowed are: {}".format(type(v), k, default_type)
                    sys.stderr.write(err+'\n')
                    raise (Exception(err))

                if isinstance(default, tuple) and len(default) == 3:
                    expected_value = default[2]
                    valid = False

                    if isinstance(v, int):
                        if expected_value[0] <= v <= expected_value[1]:
                            valid = True
                    elif isinstance(v, str):
                        if v in expected_value:
                            valid = True
                    else:
                        valid = True

                    if valid:
                        setattr(self, k, v)
                    else:
                        err = "Error parsing configuration value: Invalid value for config argument {}. Allowed are: {}".format(type(v), k, expected_value)
                        sys.stderr.write(err + '\n')
                        raise(Exception(err))

        self.CONFIG_LOADED = True


config = Config()



import json
import logging
import typing
from collections import namedtuple

from paradox.config import config as cfg
from paradox.lib import ps
from paradox.lib.utils import sanitize_key
from .core import AbstractMQTTInterface

logger = logging.getLogger('PAI').getChild(__name__)

PreparseResponse = namedtuple('preparse_response', 'topics element content')


class HomeAssistantMQTTInterface(AbstractMQTTInterface):
    def __init__(self, alarm):
        super().__init__(alarm)
        self.armed = dict()
        self.partitions = {}
        self.zones = {}
        self.device = {}
        self.detected_panel = {}

        self.first_status = True

        # TODO: Maybe homeassistant needs a separate status topic
        self.availability_topic = self.mqtt.availability_topic
        self.run_status_topic = self.mqtt.run_status_topic

    def on_connect(self, mqttc, userdata, flags, result):
        ps.subscribe(self._handle_status_update, "status_update")
        ps.subscribe(self._handle_labels_loaded, "labels_loaded")
        ps.subscribe(self._handle_panel_detected, "panel_detected")

    def _publish_run_state_sensor(self):
        configuration_topic = '{}/sensor/{}/{}/config'.format(
            cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX,
            self.detected_panel.get('serial_number', 'pai'),
            'run_status'
        )

        config = dict(
            name='Run status',
            unique_id="{}_partition_{}".format(
                self.detected_panel.get('serial_number', 'pai'),
                'run_status'
            ),
            state_topic=self.run_status_topic,
            # availability_topic=self.availability_topic,
            device=self.device
        )

        self.publish(configuration_topic, json.dumps(config), 0, cfg.MQTT_RETAIN)

    def _handle_panel_detected(self, panel):
        self.detected_panel = panel
        self.device = dict(
            manufacturer="Paradox",
            model=panel.get('model'),
            identifiers=["Paradox", panel.get('model'), panel.get('serial_number')],
            name=panel.get('model'),
            sw_version=panel.get('firmware_version')
        )

        self._publish_run_state_sensor()

    def _handle_labels_loaded(self, data):
        partitions = data.get('partition', {})
        for k, v in partitions.items():
            p_data = {}
            p_data.update(v)
            self.partitions[k] = p_data

        self.zones = data.get('zone', {})

    def _handle_status_update(self, status):
        if self.mqtt.connected:
            if 'partition' in status:
                self._process_partition_statuses(status['partition'])
            if 'zone' in status:
                self._process_zone_statuses(status['zone'])

            self.first_status = False

    def _process_partition_statuses(self, partition_statuses):
        for p_key, p_status in partition_statuses.items():
            if p_key not in self.partitions:
                continue
            partition = self.partitions[p_key]

            state_topic = '{}/{}/{}/{}/{}'.format(
                cfg.MQTT_BASE_TOPIC,
                cfg.MQTT_STATES_TOPIC,
                cfg.MQTT_PARTITION_TOPIC,
                sanitize_key(partition['key']),
                'current_state'
            )

            if self.first_status:  # For HASS auto discovery
                configuration_topic = '{}/alarm_control_panel/{}/{}/config'.format(
                    cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX,
                    self.detected_panel.get('serial_number', 'pai'),
                    sanitize_key(partition['key'])
                )
                command_topic = '{}/{}/{}/{}'.format(
                    cfg.MQTT_BASE_TOPIC,
                    cfg.MQTT_CONTROL_TOPIC,
                    cfg.MQTT_PARTITION_TOPIC,
                    sanitize_key(partition['key'])
                )
                config = dict(
                    name=partition['label'],
                    unique_id="{}_partition_{}".format(
                        self.detected_panel.get('serial_number', 'pai'),
                        partition['key']
                    ),
                    command_topic=command_topic,
                    state_topic=state_topic,
                    availability_topic=self.availability_topic,
                    device=self.device,
                    payload_disarm="disarm",
                    payload_arm_home="arm_stay",
                    payload_arm_away="arm",
                    payload_arm_night="arm_sleep"
                )

                self.publish(configuration_topic, json.dumps(config), 0, cfg.MQTT_RETAIN)

    def _process_zone_statuses(self, zone_statuses):
        for z_key, p_status in zone_statuses.items():
            if z_key not in self.zones:
                continue

            zone = self.zones[z_key]

            if self.first_status:  # For HASS auto discovery
                open_topic = '{}/{}/{}/{}/{}'.format(
                    cfg.MQTT_BASE_TOPIC,
                    cfg.MQTT_STATES_TOPIC,
                    cfg.MQTT_ZONE_TOPIC,
                    sanitize_key(zone['key']),
                    'open'
                )

                config = dict(
                    name=zone['label'],
                    unique_id="{}_zone_{}_open".format(self.detected_panel.get('serial_number', 'pai'), zone['key']),
                    state_topic=open_topic,
                    device_class="motion",
                    availability_topic=self.availability_topic,
                    payload_on="True",
                    payload_off="False",
                    device=self.device
                )

                configuration_topic = '{}/binary_sensor/{}/{}/config'.format(
                    cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX,
                    self.detected_panel.get('serial_number', 'pai'),
                    sanitize_key(zone['key'])
                )

                self.publish(configuration_topic, json.dumps(config), 0, cfg.MQTT_RETAIN)



# -*- coding: utf-8 -*-

# IP Interface

import asyncio
import binascii
import logging
import os
from typing import Awaitable

from construct import GreedyBytes, Struct, Aligned, Const, Int8ub, Bytes, Int16ul, Default

from paradox.interfaces import Interface
from paradox.config import config as cfg
from paradox.lib.async_message_manager import RAWMessageHandler
from paradox.lib.crypto import encrypt, decrypt

logger = logging.getLogger('PAI').getChild(__name__)

ip_message = Struct(
    "header" / Aligned(16, Struct(
        "sof" / Const(0xaa, Int8ub),
        "length" / Int16ul,
        "unknown0" / Default(Int8ub, 0x01),
        "flags" / Int8ub,
        "command" / Int8ub,
        "sub_command" / Default(Int8ub, 0x00),
        'unknown1' / Default(Int8ub, 0x00),
        'encrypt' / Default(Int8ub, 0x03),
    ), b'\xee'),
    "payload" / Aligned(16, GreedyBytes, b'\xee'))

ip_payload_connect_response = Struct(
    'command' / Const(0x00, Int8ub),
    'key' / Bytes(16),
    'major' / Int8ub,
    'minor' / Int8ub,
    'ip_major' / Default(Int8ub, 5),
    'ip_minor' / Default(Int8ub, 2),
    'unknown' / Default(Int8ub, 0x00),
    'unknown2' / Default(Int8ub, 0x00),
    'unknown3' / Default(Int8ub, 0x00),
    'unknown4' / Default(Int8ub, 0xee)
)


class ClientConnection():
    def __init__(self, reader, writer, alarm, key):
        self.client_writer = writer
        self.client_reader = reader
        self.alarm = alarm
        self.interface_password = key
        self.connection_key = key

    async def handle_panel_message(self, data):
        """
        Handle message from panel, which must be sent to the client
        """
        if isinstance(data, Awaitable):
            try:
                data = await data
            except asyncio.TimeoutError:
                return False

        if data is not None:
            if cfg.LOGGING_DUMP_PACKETS:
                logger.debug("PNL -> IPI (payload) {}".format(binascii.hexlify(data)))

            payload_len = len(data)

            payload = encrypt(data, self.connection_key)
            flags = 0x73

            m = ip_message.build(
                dict(header=dict(length=payload_len, unknown0=2, flags=flags, command=0), payload=payload))

            if cfg.LOGGING_DUMP_PACKETS:
                logger.debug("IPI -> APP (raw) {}".format(binascii.hexlify(m)))

            self.client_writer.write(m)

        return False  # Block further message processing

    async def handle(self):
        next_connection_key = self.connection_key
        status = 'connecting'

        while True:
            try:
                data = await asyncio.wait_for(self.client_reader.read(1000), cfg.KEEP_ALIVE_INTERVAL * 1.5)
            except asyncio.TimeoutError:
                logger.info("Timeout. Client may have disconnected")
                break
            except:
                logger.info("Client disconnected")
                break

            if not data:
                continue

            if cfg.LOGGING_DUMP_PACKETS:
                logger.debug("APP -> IPI (raw) {}".format(binascii.hexlify(data)))

            message = ip_message.parse(data)
            in_payload = message.payload

            if len(in_payload) >= 16 and message.header.flags & 0x01 != 0 and len(in_payload) % 16 == 0:
                in_payload = decrypt(in_payload, self.connection_key)[:message.header.length]

            in_payload = in_payload[:message.header.length]

            assert len(in_payload) == message.header.length, "Message payload length does not match with length in " \
                                                             "header "
            if cfg.LOGGING_DUMP_PACKETS:
                logger.debug("APP -> IPI (payload) {}".format(binascii.hexlify(in_payload)))

            force_plain_text = False
            response_code = 0x01
            out_payload = None
            if message.header.command == 0xf0:
                password = in_payload

                if password != self.interface_password:
                    logger.warning("Authentication Error")
                    break
                else:
                    logger.info("Authentication Success")

                # Generate a new key
                next_connection_key = binascii.hexlify(os.urandom(8)).upper()

                out_payload = ip_payload_connect_response.build(
                    dict(
                        key=next_connection_key,
                        major=0,
                        minor=32,
                        ip_major=1,
                        ip_minor=50,
                        unknown=113,
                        unknown2=6,
                        unknown3=0x15,
                        unknown4=44
                    ))

                flags = 0x39
            elif message.header.command == 0xf2:
                out_payload = b'\x00'
                flags = 0x39
            elif message.header.command == 0xf3:
                out_payload = binascii.unhexlify('0100000000000000000000000000000000')
                flags = 0x3b
            elif message.header.command == 0xf4:
                out_payload = b'\x01' if status == 'closing_connection' else b'\x00'
                flags = 0x39
            elif message.header.command == 0xf8:
                out_payload = b'\x01'
                flags = 0x39
            elif message.header.command == 0x00:
                response_code = 0x02
                flags = 0x73
                if in_payload[0] == 0x70 and in_payload[2] == 0x05:  # Close connection
                    out_payload = self.alarm.panel.get_message('CloseConnection').build({})
                    status = 'closing_connection'
                else:
                    try:
                        async with self.alarm.request_lock, self.alarm.busy:
                            self.alarm.connection.write(in_payload)
                    except Exception:
                        logger.exception("Send to panel")
                        break

                if in_payload[0] == 0x00:  # Just a status update
                    status = 'connected'

            else:
                logger.warning(
                    "UNKNOWN: raw: {}, payload: {}".format(binascii.hexlify(data), binascii.hexlify(in_payload)))
                continue

            if out_payload is not None:
                payload_length = len(out_payload)

                if message.header.flags & 0x08 != 0:
                    out_payload = out_payload.ljust((payload_length // 16) * 16, bytes([0xee]))

                if cfg.LOGGING_DUMP_PACKETS:
                    logger.debug("IPI -> IPI (payload) {}".format(binascii.hexlify(out_payload)))

                if message.header.flags & 0x01 != 0 and not force_plain_text:
                    out_payload = encrypt(out_payload, self.connection_key)

                m = ip_message.build(dict(
                    header=dict(
                        length=payload_length,
                        unknown0=response_code,
                        flags=flags,
                        command=message.header.command
                    ),
                    payload=out_payload))

                if cfg.LOGGING_DUMP_PACKETS:
                    logger.debug("IPI -> APP (raw) {}".format(binascii.hexlify(m)))

                self.client_writer.write(m)
                await self.client_writer.drain()

                if self.connection_key != next_connection_key:
                    self.connection_key = next_connection_key

            if status == 'closing_connection':
                break


class IPInterface(Interface):
    def __init__(self, alarm):
        super().__init__(alarm)
        self.key = cfg.IP_INTERFACE_PASSWORD
        self.addr = cfg.IP_INTERFACE_BIND_ADDRESS
        self.port = cfg.IP_INTERFACE_BIND_PORT
        self.server = None
        self.started = False
        self.client_nr = 0

    # def on_connection_lost(self):
    #     logger.error('Connection with client was lost')

    def stop(self):
        if self.server is not None:
            self.server.close()
            self.server = None
        self.started = False

    def start(self):
        logger.info("Starting %s Interface", self.name)
        self.started = True

        if not self.alarm:
            logger.info("No alarm set")
            return

        asyncio.get_event_loop().create_task(self.run())

    async def run(self):
        try:
            self.server = await asyncio.start_server(
                self.handle_client, self.addr, self.port, loop=self.alarm.work_loop
            )
            logger.info('IP Interface: serving on {}'.format(self.server.sockets[0].getsockname()))
            logger.info("IP Interface started")
        except Exception as e:
            logger.error("Failed to start IP Interface {}".format(e))

    async def handle_client(self, reader, writer):
        """
        Handle message from the remote client.

        :param reader: Socket read stream from the client
        :param writer: Socket write stream to the client
        :return: None
        """
        connection = ClientConnection(reader, writer, self.alarm, self.key)

        self.client_nr = (self.client_nr + 1) % 256
        handler_name = "%s_%d" % (self.name, self.client_nr)
        self.alarm.connection.register_handler(
            RAWMessageHandler(connection.handle_panel_message, name=handler_name)
        )

        logger.info("Client %d connected" % self.client_nr)
        await self.alarm.pause()

        await connection.handle()
        self.alarm.connection.deregister_handler(handler_name)

        logger.debug("Resuming")
        await self.alarm.resume()
        logger.info("Client %d disconnected" % self.client_nr)



# -*- coding: utf-8 -*-

import binascii
import inspect
import logging
import typing

from construct import Construct, Container, MappingError, ChecksumError

from paradox.exceptions import StatusRequestException, AuthenticationFailed
from . import parsers
from .event import event_map
from .property import property_map
from ..panel import Panel as PanelBase

logger = logging.getLogger('PAI').getChild(__name__)

ZONE_ACTIONS = dict(
    bypass={
        "flags": {
            "bypassed": True
        },
        "operation": "set",
    },
    clear_bypass={
        "flags": {
            "bypassed": True
        },
        "operation": "clear",
    },
    clear_alarm_memory={
        "flags": {
            "generated_alarm": True
        },
        "operation": "clear",
    }
)


class Panel_EVOBase(PanelBase):
    event_map = event_map
    property_map = property_map
    max_eeprom_response_data_length = 64
    status_request_addresses = parsers.RAMDataParserMap.keys()

    def get_message(self, name: str) -> Construct:
        try:
            clsmembers = dict(inspect.getmembers(parsers))
            if name in clsmembers:
                return clsmembers[name]
        except ResourceWarning:
            pass

        return super(Panel_EVOBase, self).get_message(name)

    async def dump_memory(self):
        """
        Dumps EEPROM and RAM memory to files
        :return:
        """
        await self.dump_memory_to_file('eeprom.bin', range(0, 0xffff, 64))
        await self.dump_memory_to_file('ram.bin', range(0, 59), True)

    async def dump_memory_to_file(self, file, range_, ram=False):
        mem_type = "RAM" if ram else "EEPROM"
        logger.info("Dump " + mem_type)

        packet_length = 64  # 64 is max
        with open(file, 'wb') as fh:
            for address in range_:
                args = dict(
                    address=address,
                    length=packet_length,
                    control=dict(ram_access=ram))
                logger.info("Dumping %s: address %d" % (mem_type, address))
                reply = await self.core.send_wait(
                    parsers.ReadEEPROM, args,
                    reply_expected=lambda m: m.fields.value.po.command == 0x05 and m.fields.value.address == address
                )

                if reply is None:
                    logger.error("Could not read %s: address %d" % (mem_type, address))
                    return

                data = reply.fields.value.data

                fh.write(data)

    def parse_message(self, message: bytes, direction='topanel') -> typing.Optional[Container]:
        try:
            if message is None or len(message) == 0:
                return None

            parent_parsed = super(Panel_EVOBase, self).parse_message(message, direction)
            if parent_parsed:
                return parent_parsed

            if direction == 'topanel':
                if message[0] == 0x70:
                    return parsers.CloseConnection.parse(message)
                elif message[0] == 0x00:
                    return parsers.InitializeCommunication.parse(message)
                elif message[0] == 0x30:
                    return parsers.SetTimeDate.parse(message)
                elif message[0] == 0x40:
                    return parsers.PerformPartitionAction.parse(message)
                elif message[0] == 0xd0:
                    return parsers.PerformZoneAction.parse(message)
            else:
                if message[0] >> 4 == 0x7:
                    return parsers.ErrorMessage.parse(message)
                elif message[0] >> 4 == 0x1:
                    return parsers.LoginConfirmationResponse.parse(message)
                elif message[0] >> 4 == 0x03:
                    return parsers.SetTimeDateResponse.parse(message)
                elif message[0] >> 4 == 4:  # Used for partitions and PGMs
                    return parsers.PerformActionResponse.parse(message)
                elif message[0] >> 4 == 0xd:
                    return parsers.PerformZoneActionResponse.parse(message)
                # elif message[0] == 0x50 and message[2] == 0x80:
                #     return PanelStatus.parse(message)
                # elif message[0] == 0x50 and message[2] < 0x80:
                #     return ReadEEPROM.parse(message)
                # elif message[0] >> 4 == 0x05 and message[2] == 0x80:
                #     return PanelStatusResponse[message[3]].parse(message)
                # elif message[0] >> 4 == 0x05 and message[2] < 0x80:
                elif message[0] >> 4 == 0x05:
                    return parsers.ReadEEPROMResponse.parse(message)
                # elif message[0] == 0x60 and message[2] < 0x80:
                #     return WriteEEPROM.parse(message)
                # elif message[0] >> 4 == 0x06 and message[2] < 0x80:
                #     return WriteEEPROMResponse.parse(message)
                elif message[0] >> 4 == 0x0e:
                    if message[1] == 0xff:
                        return parsers.LiveEvent.parse(message)
                    else:
                        return parsers.RequestedEvent.parse(message)

        except ChecksumError as e:
            logger.error("ChecksumError %s, message: %s" % (str(e), binascii.hexlify(message)))
        except Exception:
            logger.exception("Exception parsing message: %s" % (binascii.hexlify(message)))

        return None

    async def initialize_communication(self, reply: Container, password) -> bool:
        encoded_password = self.encode_password(password)

        raw_data = reply.fields.data + reply.checksum
        parsed = parsers.InitializeCommunication.parse(raw_data)
        parsed.fields.value.pc_password = encoded_password
        payload = parsers.InitializeCommunication.build(
            dict(fields=dict(value=parsed.fields.value)))

        logger.info("Initializing communication")
        reply = await self.core.send_wait(message=payload, reply_expected=[0x1, 0x0])

        if reply is None:
            logger.error("Initialization Failed")
            return False

        if reply.fields.value.po.command == 0x0:
            logger.error("Authentication Failed. Wrong PASSWORD. Make sure you use correct PC Password. In Babyware: "
                         "Right click on your panel -> Properties -> PC Communication (Babyware) -> PC Communication "
                         "(Babyware) Tab.")
            raise AuthenticationFailed('Wrong PASSWORD')
        else:  # command == 0x1
            if reply.fields.value.po.status.Winload_connected:
                logger.info("Authentication Success")
                return True
            else:
                logger.error("Authentication Failed")
                return False

    @staticmethod
    def _request_status_reply_check(message: Container, address: int):
        mvars = message.fields.value

        if (
                mvars.po.command == 0x5
                and mvars.control.ram_access is True
                and mvars.control.eeprom_address_bits == 0x0
                and mvars.bus_address == 0x00  # panel
                and mvars.address == address
        ):
            return True

        return False

    async def request_status(self, i: int) -> typing.Optional[Container]:
        args = dict(address=i, length=64, control=dict(ram_access=True))
        reply = await self.core.send_wait(
            parsers.ReadEEPROM, args, reply_expected=lambda m: self._request_status_reply_check(m, args['address'])
        )
        if reply is not None:
            logger.debug("Received status response: %d" % i)
            return self.handle_status(reply, parsers.RAMDataParserMap)
        else:
            raise StatusRequestException("No reply to status request: %d" % i)

    async def control_partitions(self, partitions: list, command: str) -> bool:
        """
        Control Partitions
        :param list partitions: a list of partitions
        :param str command: textual command
        :return: True if we have at least one success
        """
        args = dict(partitions=dict((i, command) for i in partitions))

        try:
            reply = await self.core.send_wait(
                parsers.PerformPartitionAction, args, reply_expected=0x04)
        except MappingError:
            logger.error('Partition command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Partition command: "%s" succeeded' % command)
        else:
            logger.info('Partition command: "%s" failed' % command)
        return reply is not None

    async def control_zones(self, zones: list, command: str) -> bool:
        """
        Control zones
        :param list zones: a list of zones
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in ZONE_ACTIONS:
            return False

        args = ZONE_ACTIONS[command].copy()
        args["zones"] = zones

        try:
            reply = await self.core.send_wait(
                parsers.PerformZoneAction, args, reply_expected=0xd)
        except MappingError:
            logger.error('Zone command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Zone command: "%s" succeeded' % command)
        else:
            logger.info('Zone command: "%s" failed' % command)
        return reply is not None

    async def control_outputs(self, outputs, command) -> bool:
        """
        Control PGM
        :param list outputs: a list of pgms
        :param str command: textual command
        :return: True if we have at least one success
        """

        args = {
            "pgms": outputs,
            "command": command
        }

        try:
            reply = await self.core.send_wait(parsers.PerformPGMAction, args, reply_expected=0x04)
        except MappingError:
            logger.error('PGM command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('PGM command: "%s" succeeded' % command)
        else:
            logger.info('PGM command: "%s" failed' % command)
        return reply is not None



from typing import Mapping, Sequence

from paradox.lib.utils import deep_merge


class ElementTypeContainer(dict):
    def filter(self, id_arr):
        remove_keys = set(self.keys()) - set(id_arr)
        for i in remove_keys:
            self.pop(i)

    def select(self, needle: str) -> Sequence[int]:
        """
        Helper function to select objects from provided dictionary

        :param needle:
        :return: Sequence[int] list of object indexes
        """
        selected = []  # type: Sequence[int]
        if needle == 'all' or needle == '0':
            selected = list(self)
        else:
            if needle.isdigit() and 0 < int(needle) < len(self):
                el = self.get(int(needle))
            else:
                el = self.get(needle)

            if el:
                if "id" not in el:
                    raise Exception("Invalid dictionary of elements provided")
                selected = [el["id"]]

        return selected

    def deep_merge(self, *dicts):
        deep_merge(self, *dicts)

    def get(self, k, default=None):
        try:
            return self.__getitem__(k)
        except KeyError:
            return default

    def __contains__(self, key):
        for k, v in self.items():
            if isinstance(v, Mapping) and "key" in v and v["key"] == key:
                return True
        return super(ElementTypeContainer, self).__contains__(self.__keytransform__(key))

    def __getitem__(self, key):
        if isinstance(key, str):
            for k, v in self.items():
                if isinstance(v, Mapping) and "key" in v and v["key"] == key:
                    return v
        return super(ElementTypeContainer, self).__getitem__(self.__keytransform__(key))

    def __setitem__(self, key, value):
        super(ElementTypeContainer, self).__setitem__(self.__keytransform__(key), value)

    def __delitem__(self, key):
        if isinstance(key, str):
            for k, v in self.items():
                if isinstance(v, Mapping) and "key" in v and v["key"] == key:
                    key = k
                    break
        super(ElementTypeContainer, self).__delitem__(self.__keytransform__(key))

    @staticmethod
    def __keytransform__(key):
        if isinstance(key, str) and key.isdigit():
            return int(key)
        return key



from .models.evo48 import Panel_EVO48
from .models.evo96 import Panel_EVO96
from .models.evo192 import Panel_EVO192
from .models.evohd import Panel_EVOHD

__all__ = ['Panel_EVO48', 'Panel_EVO96', 'Panel_EVO192', 'Panel_EVOHD']



from ..panel import Panel_EVOBase


class Panel_EVO48(Panel_EVOBase):
    mem_map = {
        "labels": {
            "zone": {
                "label_offset": 0,
                "addresses": [
                    range(0x00430, 0x00730, 0x10),  # EVO48
                ]
            },
            "pgm": {
                "label_offset": 0,
                "addresses": [range(0x07082, 0x7482, 0x20)],
                "template": {
                    "on": False,
                    "pulse": False
                }
            },
            "partition": {
                "label_offset": 0,
                "addresses": [
                    range(0x03a6b, 0x03c17, 0x6b),  # EVO48
                ]
            },
            "user": {
                "label_offset": 0,
                "addresses": [range(0x03e47, 0x04447, 0x10)]
            },
            "bus-module": {  # modules
                "label_offset": 0,
                "addresses": [
                    range(0x04e47, 0x05637, 0x10),  # EVO48
                ]
            },
            "door": {
                "label_offset": 0,
                "addresses": [range(0x0345c, 0x365c, 0x10)]
            }
        },
        "definitions": {
            "zone": {
                "addresses": [
                    range(0x01f0, 0x024e + 2, 2)  # EVO48
                ]
            },
            "partition": {
                "bit_encoded": True,
                "addresses": [
                    [0x39d8]  # All 8 partitions as bits
                ]
            }
        }
    }



import asyncio
import logging
import typing

from paradox.lib.async_message_manager import AsyncMessageManager

logger = logging.getLogger('PAI').getChild(__name__)


class ConnectionProtocol(asyncio.Protocol):
    def __init__(self, on_message: typing.Callable[[bytes], None], on_con_lost):
        self.transport = None
        self.use_variable_message_length = True
        self.on_message = on_message
        self.on_con_lost = on_con_lost

    def connection_made(self, transport):
        self.transport = transport

    def close(self):
        if self.transport:
            try:
                self.transport.close()
            except Exception:
                logger.exception("Connection transport close raised Exception")
            self.transport = None

    def send_message(self, message):
        raise NotImplementedError('This function needs to be overridden in a subclass')

    def connection_lost(self, exc):
        self.close()
        self.on_con_lost()

    def variable_message_length(self, mode):
        self.use_variable_message_length = mode


class Connection(AsyncMessageManager):
    def __init__(self, on_message: typing.Callable[[bytes], None]):
        super().__init__()
        self.connected = False
        self.connection = None  # type: ConnectionProtocol
        self.on_message = on_message

    async def connect(self):
        raise NotImplementedError("Implement in subclass")

    def write(self, data: bytes):
        if self.connected:
            self.connection.send_message(data)
        else:
            raise ConnectionError("Failed to write data to connection")

    def close(self):
        logger.info('Closing connection')

        if self.connection:
            self.connection.close()
            self.connection = None
        self.connected = False

    def variable_message_length(self, mode):
        if self.connection is not None:
            self.connection.variable_message_length(mode)



# -*- coding: utf-8 -*-

import datetime

from construct import Adapter, Array, Container, ListContainer, BitStruct, Flag, Default, Subconstruct, BitsInteger, \
    Construct, Struct, Computed, this, Bytes


class DateAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.datetime(obj[0] * 100 + obj[1], obj[2], obj[3], obj[4], obj[5], obj[6] if len(obj) > 6 else 0)

    def _encode(self, obj, context, path):
        return [obj.year / 100, obj.year % 100, obj.month, obj.day, obj.hour, obj.minute, obj.second]


class DictArray(Adapter):
    def __init__(self, count, first_index, subcon, pick_key: str = None):
        super(DictArray, self).__init__(Array(count, subcon))
        self._first_index = first_index
        self._pick_key = pick_key

    def _decode(self, obj, context, path):
        if not isinstance(obj, list):
            raise TypeError("list should be passed for decoding")

        n_obj = Container()
        for item in obj:
            v = item
            if self._pick_key:
                v = item.get(self._pick_key)
            n_obj[item._index] = v

        return n_obj

    def _encode(self, obj, context, path):
        if not isinstance(obj, dict):
            raise TypeError("dict should be passed for decoding")

        n_obj = ListContainer()

        count = self.subcon.count  # Array count

        for k in range(self._first_index, self._first_index + count):
            o = {"_index": k}
            v = obj.get(k)
            if v is not None:
                if self._pick_key:
                    v = {self._pick_key: v}
                o.update(v)
            n_obj.append(o)

        return n_obj


class EnumerationAdapter(Subconstruct):
    def __init__(self, subcon):
        super(EnumerationAdapter, self).__init__(subcon)

        def find_count(s):
            if hasattr(s, 'count'):
                return s.count
            else:
                return find_count(s.subcon)

        self.size = find_count(subcon)

    def _build(self, obj, stream, context, path):
        zones = list([i in obj for i in range(1, self.size+1)])

        return self.subcon._build(zones, stream, context, path)


def StatusFlags(count):
    return DictArray(count, 1, Struct(
                "_index" / Computed(this._index + 1),
                "flag" / Default(Flag, False)
            ), pick_key="flag")


def ZoneFlags(count, start_index_from=1):
    return DictArray(count, start_index_from, BitStruct(
        "_index" / Computed(this._index + start_index_from),
        "generated_alarm" / Default(Flag, False),
        "presently_in_alarm" / Default(Flag, False),
        "activated_entry_delay" / Default(Flag, False),
        "activated_intellizone_delay" / Default(Flag, False),
        "bypassed" / Default(Flag, False),
        "shutted_down" / Default(Flag, False),
        "tx_delay" / Default(Flag, False),
        "supervision_trouble" / Default(Flag, False),
    ))


ZoneFlagBitStruct = BitStruct(
    "generated_alarm" / Default(Flag, False),
    "presently_in_alarm" / Default(Flag, False),
    "activated_entry_delay" / Default(Flag, False),
    "activated_intellizone_delay" / Default(Flag, False),
    "bypassed" / Default(Flag, False),
    "shutted_down" / Default(Flag, False),
    "tx_delay" / Default(Flag, False),
    "supervision_trouble" / Default(Flag, False),
)


class PartitionStatus(Subconstruct):
    first2 = BitStruct(
        'fire_alarm' / Flag,
        'audible_alarm' / Flag,
        'silent_alarm' / Flag,
        'was_in_alarm' / Flag,  # is in alarm
        'arm_no_entry' / Flag,
        'arm_stay' / Flag,  # Armed in Stay mode
        'arm_away' / Flag,  # Armed in Away mode
        'arm' / Flag,  # Armed

        'lockout' / Flag,
        'programming' / Flag,
        'zone_bypassed' / Flag,
        'alarm_in_memory' / Flag,
        'trouble' / Flag,
        'entry_delay' / Flag,
        'exit_delay' / Flag,
        'ready' / Flag
    )

    last4 = BitStruct(
        'zone_supervision_trouble' / Flag,
        'zone_fire_loop_trouble' / Flag,
        'zone_low_battery_trouble' / Flag,
        'zone_tamper_trouble' / Flag,
        'voice_arming' / Flag,
        'auto_arming_engaged' / Flag,
        'fire_delay_in_progress' / Flag,
        'intellizone_engage' / Flag,

        'time_to_refresh_zone_status' / Flag,
        'panic_alarm' / Flag,
        'police_code_delay' / Flag,  # Within police code delay
        'follow_become_delay' / Flag,  # Follow become delay when is bypassed
        'remote_arming' / Flag,
        'stay_arming_auto' / Flag,  # if no entry zone is tripped
        'partition_recently_close' / Flag,
        'cancel_alarm_reporting_on_disarming' / Flag,

        'tx_delay_finished' / Flag,  # (Time Out / instant alarm)
        'auto_arm_reach' / Flag,
        'fire_delay_end' / Flag,
        'no_movement_delay_end' / Flag,
        'alarm_duration_finished' / Flag,
        'entry_delay_finished' / Flag,
        'exit_delay_finished' / Flag,
        'intellizone_delay_finished' / Flag,

        '_free0' / BitsInteger(3),
        'all_zone_closed' / Flag,  # (Bypass or not)
        'inhibit_ready' / Flag,
        'bypass_ready' / Flag,
        'force_ready' / Flag,
        'stay_instant_ready' / Flag,
    )

    def __init__(self, subcons_or_size):
        if isinstance(subcons_or_size, Construct):
            self.size = subcons_or_size.sizeof()
            subcons = subcons_or_size
        else:
            self.size = subcons_or_size
            subcons = Bytes(subcons_or_size)
        super(PartitionStatus, self).__init__(subcons)

    def _parse(self, stream, context, path):
        obj = Container()

        if self.size == 32:
            for i in range(1, 6):
                obj[i] = self.first2._parsereport(stream, context, path)
                obj[i].update(self.last4._parsereport(stream, context, path))

            obj[6] = self.first2._parsereport(stream, context, path)
        elif self.size == 16:
            obj[6] = self.last4._parsereport(stream, context, path)
            for i in range(7, 9):
                obj[i] = self.first2._parsereport(stream, context, path)
                obj[i].update(self.last4._parsereport(stream, context, path))
        else:
            raise Exception('Not supported size. Only 32 or 16')

        return obj


class EventAdapter(Adapter):
    def _decode(self, obj, context, path):
        event_group = obj[0]
        event_1_high_nibble = obj[1] >> 6
        event_2_high_nibble = (obj[1] >> 4) & 0b11
        event_1 = obj[2] + (event_1_high_nibble << 8)
        event_2 = obj[3] + (event_2_high_nibble << 8)
        partition = obj[1] & 0x0f

        return Container({
            'major': event_group,
            'minor': event_1,
            'minor2': event_2,
            'partition': partition
        })


# class CompressedEventAdapter(Adapter):
#     def _decode(self, obj, context, path):
#         day = obj[0] >> 3
#         month = (obj[0] & 0b111) << 1 | obj[1] >> 7
#         century = obj[1] & 0b1111111
#         year = century * 100 + (obj[2] >> 1)
#         hour = (obj[2] & 0b1) << 4 | obj[3] >> 4
#         minute = (obj[3] & 0b1111) << 2 | obj[4] >> 6
#
#         Container({
#             'time': datetime.datetime(year, month, day, hour, minute)
#         })
#
#     def _encode(self, obj, context, path):
#         return [obj.year / 100, obj.year % 100, obj.month, obj.day, obj.hour, obj.minute, obj.second]



# coding: utf-8

import sys
from setuptools import setup, find_packages

from paradox import VERSION

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='paradox-alarm-interface',
    version=VERSION,
    author='João Paulo Barraca',
    author_email='jpbarraca@gmail.com',
    description='Interface to Paradox Alarm Panels',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ParadoxAlarmInterface/pai',
    download_url=f'https://github.com/ParadoxAlarmInterface/pai/archive/{VERSION}.tar.gz',
    packages=find_packages(exclude=['tests', 'tests.*', 'config.*', 'docs.*']),
    install_requires=[
        'construct~=2.9.43',
        'argparse>=1.4.0'
    ],
    python_requires='>=3.6',
    setup_requires=['wheel'] + pytest_runner,
    tests_require=[
        'pytest',
        'pytest-asyncio',
        'pytest-env',
        'pytest-mock',
        'mock'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
        'Operating System :: OS Independent',
    ],
    extras_require={
        'YAML':  ['pyyaml>=5.2.0'],
        'Serial': ['pyserial-asyncio>=0.4'],
        'IP': ['requests>=2.20.0'],
        'MQTT': ['paho_mqtt>=1.5.0'],
        'Pushbullet': [
            'pushbullet.py>=0.11.0',
            'ws4py>=0.4.2'
        ],
        'Pushover': ['chump>=1.6.0'],
        'Signal': ['pygobject>=3.20.0', 'pydbus>=0.6.0', 'gi>=1.2']
    },
    entry_points={
        'console_scripts': [
            'ip150-connection-decrypt = paradox.console_scripts.ip150_connection_decrypt:main [YAML]',
            'pai-service = paradox.console_scripts.pai_run:main'
        ]
    },
    license='EPL'
)



#!/usr/bin/env python3

__author__ = "João Paulo Barraca, Jevgeni Kiski"
__copyright__ = "Copyright 2018-2019, João Paulo Barraca"
__credits__ = ["Tihomir Heidelberg", "Louis Rossouw"]
__license__ = "EPL"
__maintainer__ = "João Paulo Barraca"
__email__ = "jpbarraca@gmail.com"
__status__ = "Beta"

import argparse
import sys

if sys.version_info[0] < 3 or sys.version_info[1] < 6:
    print("You are using Python %s.%s, but PAI requires at least Python 3.6" % (sys.version_info[0], sys.version_info[1])) 
    sys.exit(-1)

def main():
    from paradox.main import main
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", type=str, default=None,
                        help="specify path to an alternative configuration file")

    args = parser.parse_args()

    main(args)


if __name__ == '__main__':
    from paradox.lib import help
    try:
        main()
    except ImportError as error:
        help.import_error_help(error)



import asyncio
import logging
from typing import Awaitable, Optional, Callable

from construct import Container

logger = logging.getLogger('PAI').getChild(__name__)


class FutureMessageHandler(asyncio.Future):
    def __init__(self, check_fn=None, loop=None, name=None):
        super(FutureMessageHandler, self).__init__(loop=loop)
        self.persistent = False
        self._check_fn = check_fn
        self.name = name if name is not None else self.__class__.__name__

    def can_handle(self, message: Container) -> bool:
        if isinstance(self._check_fn, Callable):
            return self._check_fn(message)
        return True


class RawFutureMessageHandler(FutureMessageHandler):
    pass


class MessageHandler:
    # handle: Callable

    def __init__(self, callback, name=None):
        self.persistent = False
        self.handle = callback
        self.name = name if name is not None else self.__class__.__name__

    def can_handle(self, message: Container) -> bool:
        return True


class RAWMessageHandler(MessageHandler):
    def __init__(self, callback, name=None):
        super(RAWMessageHandler, self).__init__(callback, name)
        self.persistent = True
        self.name = name if name is not None else self.__class__.__name__


class EventMessageHandler(MessageHandler):
    def __init__(self, callback, name=None):
        super(EventMessageHandler, self).__init__(callback, name)
        self.persistent = True
        self.name = name if name is not None else self.__class__.__name__

    def can_handle(self, message: Container) -> bool:
        values = message.fields.value
        return values.po.command == 0xe and (not hasattr(values, "event_source") or values.event_source == 0xff)


class ErrorMessageHandler(MessageHandler):
    def __init__(self, callback, name=None):
        super(ErrorMessageHandler, self).__init__(callback, name)
        self.persistent = True
        self.name = name if name is not None else self.__class__.__name__

    def can_handle(self, message: Container) -> bool:
        return message.fields.value.po.command == 0x7 and hasattr(message.fields.value, "message")


class AsyncMessageManager:
    # handlers: List[MessageHandler]

    def __init__(self, loop=None):
        super(AsyncMessageManager, self).__init__()

        if not loop:
            loop = asyncio.get_event_loop()
        self.loop = loop

        self.handlers = []
        self.raw_handlers = []

    async def wait_for_message(self, check_fn=None, timeout=2, raw=False) -> Optional[Container]:
        if raw:
            future = RawFutureMessageHandler(loop=self.loop)
        else:
            future = FutureMessageHandler(check_fn, loop=self.loop)

        self.register_handler(future)

        return await asyncio.wait_for(future, timeout=timeout, loop=self.loop)

    def register_handler(self, handler):
        if isinstance(handler, (RAWMessageHandler, RawFutureMessageHandler,)):
            self.raw_handlers.append(handler)
        else:
            self.handlers.append(handler)

    def deregister_handler(self, name):
        self.handlers = list(filter(lambda x: x.name != name, self.handlers))
        self.raw_handlers = list(filter(lambda x: x.name != name, self.raw_handlers))

    def schedule_message_handling(self, message: Container):
        return self.loop.create_task(self._handle_message(message))

    def schedule_raw_message_handling(self, message):
        return self.loop.create_task(self._handle_raw_message(message))

    async def _handle_raw_message(self, message: Container):
        self.raw_handlers = self._cleanup_handlers(self.raw_handlers)

        for handler in self.raw_handlers:
            if not handler.can_handle(message):
                continue

            if not handler.persistent:
                self.raw_handlers = list(filter(lambda x: x != handler, self.raw_handlers))

            await self._handle(handler, message)

    async def _handle_message(self, message: Container):
        handler = None

        self.handlers = self._cleanup_handlers(self.handlers)

        for h in self.handlers:
            try:
                if h.can_handle(message):
                    handler = h
                    break
            except Exception as e:
                logger.exception("Exception caught during message handling")

        if handler:
            if not handler.persistent:
                self.handlers = list(filter(lambda x: x != handler, self.handlers))

            return await self._handle(handler, message)
        else:
            logger.error("No handler for message {}\nDetail: {}".format(message.fields.value.po.command, message))

    @staticmethod
    async def _handle(handler, message):
        if isinstance(handler, asyncio.Future):
            handler.set_result(message)
            return await handler
        else:
            result = handler.handle(message)
            if isinstance(result, Awaitable):
                return await result
            else:
                return result

    @staticmethod
    def _cleanup_handlers(handlers):
        return list(filter(lambda x: not (isinstance(x, asyncio.Future) and x.done()),
                    handlers))  # remove timeouted and done futures



from paradox.event import EventLevel
from time import time


def _toggle_provider(old):
    return not old


def _timestamp_provider(old):
    return int(time())


event_opt1 = {1: dict(type='output', id=1),
              2: dict(type='output', id=2),
              3: dict(type='output', id=3),
              4: dict(type='output', id=4),
              5: dict(type='output', id=5),
              6: dict(type='output', id=6),
              7: dict(type='output', id=7),
              8: dict(type='output', id=8),
              9: dict(type='output', id=9),
              10: dict(type='output', id=10),
              11: dict(type='output', id=11),
              12: dict(type='output', id=12),
              13: dict(type='output', id=13),
              14: dict(type='output', id=14),
              15: dict(type='output', id=15),
              16: dict(type='output', id=16),
              17: dict(type='repeater', id=1),
              18: dict(type='repeater', id=2),
              19: dict(type='keypad', id=1),
              20: dict(type='keypad', id=2),
              21: dict(type='keypad', id=3),
              22: dict(type='keypad', id=4),
              27: dict(type='siren', id=1),
              28: dict(type='siren', id=2),
              29: dict(type='siren', id=3),
              30: dict(type='siren', id=4),
              99: dict()
              }

event_map = {0: dict(
    level=EventLevel.DEBUG, change=dict(open=False), type='zone', message='Zone {label} OK'),
    1: dict(
        level=EventLevel.DEBUG, change=dict(open=True), type='zone', message='Zone {label} open'),
    2: dict(
        level=EventLevel.CRITICAL, type='partition',
        sub={0: dict(message='N/A'),
             1: dict(message='N/A'),
             2: dict(change=dict(silent_alarm=True, alarm=True), tags=['alarm', 'silent_alarm'],
                     message='Partition {label} silent alarm active'),
             3: dict(change=dict(audible_alarm=True, alarm=True), tags=['alarm', 'buzzer_alarm'],
                     message='Partition {label} buzzer alarm active'),
             4: dict(change=dict(audible_alarm=True, alarm=True), tags=['alarm', 'steady_alarm'],
                     message='Partition {label} steady alarm active'),
             5: dict(change=dict(audible_alarm=True, alarm=True), tags=['alarm', 'pulse_alarm'],
                     message='Partition {label} pulse alarm active'),
             6: dict(change=dict(strobe_alarm=True, alarm=True), tags=['alarm', 'strobe'],
                     message='Partition {label} strobe active'),
             7: dict(level=EventLevel.INFO, change=dict(alarm=False), message='Partition {label} alarm stopped'),
             8: dict(level=EventLevel.DEBUG, change=dict(squawk=True), id=1, message='Squawk ON'),
             9: dict(level=EventLevel.DEBUG, change=dict(squawk=False), id=1, message='Squawk OFF'),
             10: dict(level=EventLevel.DEBUG, message='Ground Start'),
             11: dict(level=EventLevel.INFO, change=dict(arm=False, arm_sleep=False, arm_stay=False, alarm=False),
                      tags=["disarm"], message='Partition {label} disarmed'),
             12: dict(level=EventLevel.INFO, change=dict(arm=True), tags=["arm"], message='Partition {label} armed'),
             13: dict(level=EventLevel.DEBUG, change=dict(entry_delay=True),
                      message='Partition {label} entry delay started'),
             14: dict(level=EventLevel.DEBUG, change=dict(exit_delay=True),
                      message='Partition {label} exit delay started'),
             15: dict(level=EventLevel.DEBUG, message='Pre-alarm delay'),
             16: dict(level=EventLevel.DEBUG, message='Report confirmation'),
             99: dict(level=EventLevel.DEBUG, message='Any partition status event')}
    ),
    3: dict(
        level=EventLevel.DEBUG, type='system',
        sub={0: dict(change=dict(bell_activated=False), message='Bell OFF'),
             1: dict(change=dict(bell_activated=True), message='Bell ON'),
             2: dict(message='Bell squawk arm'),
             3: dict(message='Bell squawk disarm'),
             99: dict(message='Any bell status event')}
    ),
    # 5: 'Non-Reportable Event',
    6: dict(
        level=EventLevel.DEBUG, message='Non-reportable event',
        sub={0: dict(level=EventLevel.CRITICAL, type='system', tags=['trouble'], message='Telephone line trouble'),
             1: dict(type='system', message='[ENTER]/[CLEAR]/[POWER] key was pressed'),
             2: dict(message='N/A'),
             3: dict(level=EventLevel.INFO, change=dict(arm_stay=True, arm=True), type='partition', tags=['arm'],
                     message='Arm in stay mode'),
             4: dict(level=EventLevel.INFO, change=dict(arm_sleep=True, arm=True), type='partition', tags=['arm'],
                     message='Arm in sleep mode'),
             5: dict(level=EventLevel.INFO, change=dict(arm_force=True, arm=True), type='partition', tags=['arm'],
                     message='Arm in force mode'),
             6: dict(level=EventLevel.INFO, change=dict(arm_stay=True, arm=True), type='partition', tags=['arm'],
                     message='Full arm when armed in stay mode'),
             7: dict(level=EventLevel.CRITICAL, tags=['trouble'],
                     change=dict(computer_fail_to_communicate_trouble=True), type='system',
                     message='PC fail to communicate'),
             8: dict(type='system', message='Utility Key 1 pressed (keys [1] and [2])'),
             9: dict(type='system', message='Utility Key 2 pressed (keys [4] and [5])'),
             10: dict(type='system', message='Utility Key 3 pressed (keys [7] and [8])'),
             11: dict(type='system', message='Utility Key 4 pressed (keys [2] and [3])'),
             12: dict(type='system', message='Utility Key 5 pressed (keys [5] and [6])'),
             13: dict(type='system', message='Utility Key 6 pressed (keys [8] and [9])'),
             14: dict(level=EventLevel.CRITICAL, type='system', tags=['alarm'], message='Tamper generated alarm'),
             15: dict(level=EventLevel.CRITICAL, type='system', tags=['supervision', 'alarm'],
                      message='Supervision loss generated alarm'),
             16: dict(message='N/A'),
             17: dict(message='N/A'),
             18: dict(message='N/A'),
             19: dict(message='N/A'),
             20: dict(level=EventLevel.INFO, change=dict(arm=True), type='partition',
                      message='Full arm when armed in sleep mode'),
             21: dict(type='firmware', message='Firmware upgrade'),
             22: dict(message='N/A'),
             23: dict(change=dict(stayd_mode_active=True), type='partition', message='StayD mode activated'),
             24: dict(change=dict(stayd_mode_active=False), type='partition', message='StayD mode deactivated'),
             25: dict(type='system', message='IP Registration status change'),
             26: dict(type='system', message='GPRS Registration status change'),
             27: dict(level=EventLevel.INFO, type='partition', message='Armed with trouble(s)'),
             28: dict(level=EventLevel.WARN, type='system', tags=['trouble', 'supervision'],
                      message='Supervision alert'),
             29: dict(level=EventLevel.INFO, type='system', tags=['trouble', 'restore', 'supervision'],
                      message='Supervision alert restore'),
             30: dict(level=EventLevel.WARN, type='partition', message='Armed with remote with low battery'),
             99: dict(message='Any non-reportable event')}
    ),
    # 7: 'PGM Activation',
    8: dict(
        level=EventLevel.DEBUG, type='user', change=dict(button_b=_timestamp_provider),
        message='Button B pressed on remote of {@user}'),
    9: dict(
        level=EventLevel.DEBUG, type='user', change=dict(button_c=_timestamp_provider),
        message='Button C pressed on remote of {@user}'),
    10: dict(
        level=EventLevel.DEBUG, type='user', change=dict(button_d=_timestamp_provider),
        message='Button D pressed on remote of {@user}'),
    11: dict(
        level=EventLevel.DEBUG, type='user', change=dict(button_e=_timestamp_provider),
        message='Button E pressed on remote of {@user}'),
    12: dict(
        level=EventLevel.DEBUG, type='user', message='Cold start wireless zone {label}'),
    13: dict(
        level=EventLevel.DEBUG, type='user', message='Cold start wireless module {label}'),
    14: dict(
        level=EventLevel.DEBUG, type='zone', change=dict(bypassed=_toggle_provider),
        message='Bypass programming on zone {label}'),
    15: dict(
        level=EventLevel.DEBUG, type='partition', message='User code activated output {label}'),
    16: dict(
        level=EventLevel.DEBUG, type='zone', message='Wireless smoke maintenance signal'),
    17: dict(
        level=EventLevel.DEBUG, type='zone', message='Delay zone {label} alarm transmission'),
    18: dict(
        level=EventLevel.WARN, type='zone', message='Zone {label} signal strength weak 1'),
    19: dict(
        level=EventLevel.WARN, type='zone', message='Zone {label} signal strength weak 2'),
    20: dict(
        level=EventLevel.WARN, type='zone', message='Zone {label} signal strength weak 3'),
    21: dict(
        level=EventLevel.WARN, type='zone', message='Zone {label} signal strength weak 4'),
    22: dict(
        level=EventLevel.DEBUG, type='user', change=dict(button_5=_timestamp_provider),
        message='Button 5 pressed on remote of {@user}'),
    23: dict(
        level=EventLevel.DEBUG, type='user', change=dict(button_6=_timestamp_provider),
        message='Button 6 pressed on remote of {@user}'),
    24: dict(
        level=EventLevel.WARN, type='zone', change=dict(fire_delay=True), message='Fire delay started'),
    # 25: 'N/A',
    26: dict(
        level=EventLevel.DEBUG, type='system', message='Software access',
        sub={0: dict(message='Non-valid source ID'),
             1: dict(message='WinLoad direct'),
             2: dict(message='WinLoad through IP module'),
             3: dict(message='WinLoad through GSM module'),
             4: dict(message='WinLoad through modem'),
             9: dict(message='IP100 direct'),
             10: dict(message='VDMP3 direct'),
             11: dict(message='Voice through GSM module'),
             12: dict(message='Remote access'),
             13: dict(message='SMS through GSM module'),
             99: dict(message='Any software access')}
    ),
    27: dict(
        level=EventLevel.DEBUG, type='bus', message='Bus module event',
        sub={0: dict(message='A bus module was added'),
             1: dict(message='A bus module was removed'),
             2: dict(message='2-way RF Module Communication Failure'),
             3: dict(message='2-way RF Module Communication Restored')}
    ),
    28: dict(
        level=EventLevel.INFO, type='zone', message='StayD pass acknowledged for {label}'),
    29: dict(
        level=EventLevel.INFO, type='user', tags=['arm'], message='Arming by user {label}'),
    30: dict(
        level=EventLevel.INFO, type='special', message="Special arming",
        sub={0: dict(tags=['arm'], message='Auto-arming (on time/no movement)'),
             1: dict(message='Late to close'),
             2: dict(tags=['arm'], message='No movement arming'),
             3: dict(tags=['arm'], message='Partial arming'),
             4: dict(tags=['arm'], message='Quick arming'),
             5: dict(tags=['arm'], message='Arming through WinLoad'),
             6: dict(tags=['arm'], message='Arming with keyswitch'),
             99: dict(tags=['arm'], message='Any special arming')}
    ),
    31: dict(
        level=EventLevel.INFO, type='user', tags=['disarm'], message='Disarming by user {label}'),
    32: dict(
        level=EventLevel.INFO, type='user', tags=['disarm', 'after_alarm'],
        message='Disarming after alarm by user {label}'),
    33: dict(
        level=EventLevel.INFO, type='user', tags=['alarm', 'cancel'], message='Alarm cancelled by user {label}'),
    34: dict(
        level=EventLevel.WARN, type='special',
        sub={0: dict(message='Auto-arm cancelled (on time/no movement)'),
             1: dict(level=EventLevel.INFO, tags=['disarm'], message='Disarming through WinLoad'),
             2: dict(level=EventLevel.INFO, tags=['disarm', 'after_alarm'],
                     message='Disarming through WinLoad after alarm'),
             3: dict(level=EventLevel.INFO, tags=['alarm', 'cancel'], message='Alarm cancelled through WinLoad'),
             4: dict(level=EventLevel.INFO, tags=['alarm', 'cancel'], message='Paramedical alarm cancelled'),
             5: dict(level=EventLevel.DEBUG, tags=['disarm'], message='Disarm with keyswitch'),
             6: dict(level=EventLevel.DEBUG, tags=['disarm', 'after_alarm'],
                     message='Disarm with keyswitch after an alarm'),
             7: dict(level=EventLevel.INFO, tags=['alarm', 'cancel'], message='Alarm cancelled with keyswitch'),
             99: dict(level=EventLevel.INFO, tags=['disarm'], message='Any special disarming')}
    ),
    35: dict(
        level=EventLevel.INFO, change=dict(bypassed=_toggle_provider), type='zone', tags=['bypass'],
        message='Zone {label} bypass toggled'),
    36: dict(
        level=EventLevel.CRITICAL, change=dict(alarm=True), type='zone', tags=['alarm', 'trigger'],
        message='Zone {label} in alarm'),
    37: dict(
        level=EventLevel.CRITICAL, change=dict(fire_alarm=True), type='zone', tags=['alarm', 'fire', 'trigger'],
        message='Zone {label} in fire alarm'),
    38: dict(
        level=EventLevel.CRITICAL, change=dict(alarm=False), type='zone', tags=['alarm', 'restore'],
        message='Zone {label} alarm restore'),
    39: dict(
        level=EventLevel.CRITICAL, change=dict(fire_alarm=False), type='zone', tags=['alarm', 'fire', 'restore'],
        message='Zone {label} fire alarm restore'),
    40: dict(
        level=EventLevel.CRITICAL, type='special', tags=['alarm', 'trigger'], message="Special alarm",
        sub={0: dict(message='Panic non-medical emergency'),
             1: dict(message='Panic medical'),
             2: dict(message='Panic fire'),
             3: dict(message='Recent closing'),
             4: dict(message='Global shutdown'),
             5: dict(message='Duress alarm'),
             6: dict(message='Keypad lockout'),
             99: dict(message='Any special alarm event')}
    ),
    41: dict(
        level=EventLevel.INFO, change=dict(shutdown=True), type='zone', message='Zone {label} shutdown'),
    42: dict(
        level=EventLevel.WARN, change=dict(zone_tamper_trouble=True), tags=['tamper'], type='zone',
        message='Zone {label} tampered'),
    43: dict(
        level=EventLevel.INFO, change=dict(zone_tamper_trouble=False), tags=['tamper', 'restore'], type='zone',
        message='Zone {label} tamper restore'),
    44: dict(
        level=EventLevel.WARN, type='system', tags=['trouble'],
        sub={0: dict(message='N/A'),
             1: dict(message='AC failure'),
             2: dict(message='Battery failure'),
             3: dict(message='Auxiliary current overload'),
             4: dict(message='Bell current overload'),
             5: dict(message='Bell disconnected'),
             6: dict(change=dict(timer_loss_trouble=True), message='Clock loss'),
             7: dict(change=dict(fire_loop_trouble=True), message='Fire loop trouble'),
             8: dict(message='Fail to communicate to monitoring station telephone #1'),
             9: dict(message='Fail to communicate to monitoring station telephone #2'),
             11: dict(message='Fail to communicate to voice report'),
             12: dict(message='RF jamming'),
             13: dict(message='GSM RF jamming'),
             14: dict(message='GSM no service'),
             15: dict(message='GSM supervision lost'),
             16: dict(message='Fail To Communicate IP Receiver 1 (GPRS)'),
             17: dict(message='Fail To Communicate IP Receiver 2 (GPRS)'),
             18: dict(message='IP Module No Service'),
             19: dict(message='IP Module Supervision Loss'),
             20: dict(message='Fail To Communicate IP Receiver 1 (IP)'),
             21: dict(message='Fail To Communicate IP Receiver 2 (IP)'),
             99: dict(message='Any new trouble event')}
    ),
    45: dict(
        level=EventLevel.INFO, type='system', tags=['trouble', 'restore'],
        sub={0: dict(message='Telephone line restore'),
             1: dict(message='AC failure restore'),
             2: dict(message='Battery failure restore'),
             3: dict(message='Auxiliary current overload restore'),
             4: dict(message='Bell current overload restore'),
             5: dict(message='Bell disconnected restore'),
             6: dict(message='Clock loss restore'),
             7: dict(message='Fire loop trouble restore'),
             8: dict(message='Fail to communicate to monitoring station telephone #1 restore'),
             9: dict(message='Fail to communicate to monitoring station telephone #2 restore'),
             11: dict(message='Fail to communicate to voice report restore'),
             12: dict(message='RF jamming restore'),
             13: dict(message='GSM RF jamming restore'),
             14: dict(message='GSM no service restore'),
             15: dict(message='GSM supervision lost restore'),
             16: dict(message='Fail To Communicate IP Receiver 1 (GPRS) restore'),
             17: dict(message='Fail To Communicate IP Receiver 2 (GPRS) restore'),
             18: dict(message='IP Module No Service restore'),
             19: dict(message='IP Module Supervision Loss restore'),
             20: dict(message='Fail To Communicate IP Receiver 1 (IP) restore'),
             21: dict(message='Fail To Communicate IP Receiver 2 (IP) restore'),
             99: dict(message='Any trouble event restore')}
    ),
    46: dict(
        level=EventLevel.WARN, type='module', tags=['trouble'],
        sub={0: dict(message='Bus / EBus / Wireless module communication fault'),
             1: dict(message='Tamper trouble'),
             2: dict(message='Power fail'),
             3: dict(message='Battery failure'),
             99: dict(message='Any bus module new trouble event')}
    ),
    47: dict(
        level=EventLevel.INFO, type='system', tags=['trouble', 'restore'],
        sub={0: dict(message='Bus / EBus / Wireless module communication fault restore'),
             1: dict(message='Tamper trouble restore'),
             2: dict(message='Power fail restore'),
             3: dict(message='Battery failure restore'),
             99: dict(message='Any bus module trouble restored event')}
    ),
    48: dict(
        level=EventLevel.INFO, type='system',
        sub={0: dict(message='System power up'),
             1: dict(message='Reporting test'),
             2: dict(message='Software log on'),
             3: dict(message='Software log off'),
             4: dict(message='Installer in programming mode'),
             5: dict(message='Installer exited programming mode'),
             6: dict(message='Maintenance in programming mode'),
             7: dict(message='Maintenance exited programming mode'),
             8: dict(message='Closing delinquency delay elapsed'),
             99: dict(message='Any special event')}
    ),
    49: dict(
        level=EventLevel.WARN, changes=dict(zone_rf_low_battery_trouble=True),
        type='zone', tags=['trouble'], message='Low battery on zone {label}'),
    50: dict(
        level=EventLevel.INFO, changes=dict(zone_rf_low_battery_trouble=False),
        type='zone', tags=['trouble', 'restore'], message='Low battery on zone {label} restore'),
    51: dict(
        level=EventLevel.WARN, changes=dict(zone_supervision_trouble=True),
        type='zone', tags=['trouble', 'supervision'], message='Zone {label} supervision trouble'),
    52: dict(
        level=EventLevel.INFO, changes=dict(zone_supervision_trouble=False),
        type='zone', tags=['trouble', 'supervision', 'restore'], message='Zone {label} supervision restore'),
    53: dict(
        level=EventLevel.WARN, change=dict(module_supervision_trouble=True), type='module',
        tags=['trouble', 'supervision'],
        message='Wireless module {label} supervision trouble', sub=event_opt1),
    54: dict(
        level=EventLevel.INFO, change=dict(module_supervision_trouble=False), type='module',
        tags=['trouble', 'supervision', 'restore'],
        message='Wireless module {label} supervision restore', sub=event_opt1),
    55: dict(
        level=EventLevel.WARN, change=dict(module_tamper_trouble=True), type='module', tags=['trouble', 'tamper'],
        message='Wireless module {label} tamper trouble', sub=event_opt1),
    56: dict(
        level=EventLevel.INFO, change=dict(module_tamper_trouble=False), type='module',
        tags=['trouble', 'tamper', 'restore'],
        message='Wireless module {label} tamper restore'),
    57: dict(
        level=EventLevel.CRITICAL, type='user', tags=['alarm', 'medic', 'trigger'],
        message='Non-medical alarm (paramedic) by {label}'),
    58: dict(
        level=EventLevel.DEBUG, type='zone', message='Zone {label} forced'),
    59: dict(
        level=EventLevel.DEBUG, type='zone', message='Zone {label} included'),
    64: dict(
        level=EventLevel.DEBUG, type='system',
        sub={0: dict(message='Follow Arm LED status'),
             1: dict(message='PGM pulse fast in alarm'),
             2: dict(message='PGM pulse fast in exit delay below 10 sec.'),
             3: dict(message='PGM pulse slow in exit delay over 10 sec.'),
             4: dict(message='PGM steady ON if armed'),
             5: dict(message='PGM OFF if disarmed')}
    )
}

_labels = {
    'zone': 'Zone Number {label}',
    'pgm': 'PGM Number {label}',
    'user': 'User {label}',
    'keypad': 'Wireless Keypad {label}',
    'repeater': 'Wireless Repeater {label}',
    'siren': 'Wireless Siren {label}'
}



from ..panel import Panel_EVOBase


class Panel_EVO192(Panel_EVOBase):
    mem_map = {
        "labels": {
            "zone": {
                "label_offset": 0, "addresses": [
                    range(0x00430, 0x00730, 0x10),  # EVO48
                    range(0x00730, 0x00a30, 0x10),  # EVO96 = EVO48 + 48 zones
                    range(0x062f7, 0x068f7, 0x10)  # EVO192 = EVO96 + 96 zones
                ]},
            "pgm": {
                "label_offset": 0, "addresses": [range(0x07082, 0x7482, 0x20)], "template": {
                    "on": False,
                    "pulse": False}
            },
            "partition": {
                "label_offset": 0, "addresses": [
                    range(0x03a6b, 0x03c17, 0x6b),  # EVO48
                    range(0x03c17, 0x03dc3, 0x6b)  # EVO96 & EVO192 = EVO48 + 4 partitions
                ]},
            "user": {
                "label_offset": 0, "addresses": [range(0x03e47, 0x04e47, 0x10)]
            },
            "bus-module": {  # modules
                "label_offset": 0, "addresses": [
                    range(0x04e47, 0x05637, 0x10),  # EVO48
                    range(0x05637, 0x05e27, 0x10)  # EVO96 & EVO192 = EVO48 + 127 modules
                ]
            },
            "door": {
                "label_offset": 0, "addresses": [range(0x0345c, 0x365c, 0x10)]
            }
        },
        "definitions": {
            "zone": {
                "addresses": [
                    range(0x01f0, 0x02ae + 2, 2),  # EVO96
                    range(0x60b7, 0x6175 + 2, 2)  # EVO192
                ]
            },
            "partition": {
                "bit_encoded": True,
                "addresses": [
                    [0x39d8]  # All 8 partitions as bits
                ]
            }
        }
    }



from construct import Struct, Aligned, Const, Int8ub, Bytes, Int16ub, Int16ul, Default, Enum, GreedyBytes

from paradox.hardware.common import HexInt

ip_message = Struct(
        "header" / Aligned(16, Struct(
            "sof" / Const(0xaa, Int8ub), 
            "length" / Int16ul,
            "unknown0" / Default(Int8ub, 0x01),
            "flags" / Int8ub,
            "command" / Int8ub,
            "sub_command" / Default(Int8ub, 0x00),
            'unknown1' / Default(Int8ub, 0x0a),
            'encrypt' / Default(Int8ub, 0x00),
        ), b'\xee'),    
        "payload" / Aligned(16, GreedyBytes, b'\xee')
      )

ip_payload_connect_response = Struct(
    'login_status' / Enum(Int8ub,
                          success=0x00,
                          invalid_password=0x01,
                          user_already_connected=0x02,
                          user_already_connected1=0x04),
    'key' / Bytes(16),
    'hardware_version' / Int16ub,
    'ip_firmware_major' / Default(HexInt, 5),
    'ip_firmware_minor' / Default(HexInt, 2),
    'ip_module_serial' / Bytes(4),
)
